/*
	New Sherlock preprocessor -- Main routine

	PUBLIC DOMAIN SOFTWARE -- See below.

	source:  spp.c
	started: October 7, 1985

	Version 1.0:  July 22, 1988
	Version 1.0a: November 1, 1988 bug fix in sys.c
	Version 1.0b: November 3, 1988 bug fix in spp.c, def.c
	Version 1.1:
		February 10, 1989 bug fix in spp.c and dir.c
		February 16, 1989 is_sherlock() added.
	Version 1.1A: February 25, 1989: bug fix in so_disable().
	Version 1.2:  February 27, 1989: sl_sbout support().
	Version 1.4:  June 22, June 26 and June 27, 1989:
		
		o Added support for INCLUDE environment variable in main().
		o sysnext() changed to eliminate all backslash-newlines.
		o tok.c, def.c changed to eliminate backslash-newline logic.
		  (eliminated with #ifdef BACKSLASH_NEWLINE)
		o fixed comment bug in tok.c.
		o fixed run-on macro expansion bug in def.c.

	Version 1.5: August 4, 1989
		o -x option disables single-line comments.
		o Added slc_flag as a sentinal to single-line comments.
		o Enabled single-line comment.
		o Added single() support for control-c interrupt.

	Version 1.7, 1991
		o This was the first Public Domain version.
		o Bug fix to is_reserved.
		o Bug fix to st_init.
		o Bug fixes in sem.c


	PUBLIC DOMAIN SOFTWARE

	Sherlock, including the SPP, SDEL and SDIF programs, was placed in
	the public domain on June 15, 1991, by its author,

		Edward K. Ream
		166 North Prospect Ave.
		Madison, WI 53705.
		(608) 257-0802

	Sherlock may be used for any commercial or non-commercial purpose.


	DISCLAIMER OF WARRANTIES

	Edward K. Ream (Ream) specifically disclaims all warranties,
	expressed or implied, with respect to this computer software,
	including but not limited to implied warranties of merchantability
	and fitness for a particular purpose.  In no event shall Ream be
	liable for any loss of profit or any commercial damage, including
	but not limited to special, incidental consequential or other damages.
	
*/

#include "spp.h"

/* Declare local routines. */
static void	synonyms(void);
static void	set_syn	(char *, char *);

#define SIGNON "SPP v1.7: June 15, 1991"
#define DEBUG_WARNING "---debug version---"

#ifdef SHERLOCK
#define USAGE1 "usage: spp in out [options] ++/--routine\n\n"
#else
#define USAGE1 "usage: spp in out [options]\n\n"
#endif

#define USAGE2  "-d id=value  #define id value\n"
#define USAGE3	"-f <file>    use synonym file\n"
#define USAGE4  "-i           insert include \"sl.h\" at start of file\n"
#define USAGE5  "-n           allow nested comments\n"
#define USAGE6  "-o           use usr output routines instead of printf\n"
#define USAGE7  "-s <path>    set search path for #include's\n"
#define USAGE8	"-t           insert tick-style macros\n"
#define USAGE9  "-u id        #undef id\n"
#define USAGE10 "-x           do not allow single-line comments\n"
#define USAGE11 "-?           print the version number and exit\n"

void
main(int argc, char **argv)
{
	char *in = NULL, *out = NULL;
	char *arg;
	char *def;
	char *p, *p1;
	char path_buf[400];

	/* These two calls MUST come before any others. */
	SL_INIT();
	SL_PARSE(argc, argv, "++", "--");

	/*
		The call to mst_init() MUST be placed here so we can define
		macros using the -d option.

		The call to mst2_init() must FOLLOW the gathering of
		command line arguments so that the __line__ and __file__
		macros may be disabled using the -u option.
	*/
	mst_init();
	st_init();

	/* Allow user to abort. */
	syscsts();

	TICK("main");

	/* Always put out the sign on message. */
#ifdef SHERLOCK
	printf("%s%s\n", SIGNON, DEBUG_WARNING);
#else
	printf("%s\n", SIGNON);
#endif


	/* Make first test for correct command line. */
	if (argc == 2 && str_eq(argv[1], "-?")) {
		exit(BAD_EXIT);
	}
	if (argc < 3) {
		printf("%s%s%s%s", USAGE1,  USAGE2,  USAGE3,  USAGE4);
		printf("%s%s%s%s", USAGE5,  USAGE6,  USAGE7,  USAGE8);
		printf("%s%s%s", USAGE9, USAGE10, USAGE11);
		exit(BAD_EXIT);
	}

	/* Process all the arguments on the command line. */
	argc--;
	argv++;
	while (argc-- > 0) {
		arg = *argv++;

		if(str_eq(arg, "-d")) {
			/* Define a variable. */
			if (argc--) {
				arg = *argv++;
				/* Scan for an optional equal sign. */
				for (def = arg;	*def; def++) {
					if (*def == '=') {
						*def = '\0';
						def++;
						break;
					}
				}
				mst_enter(arg, def, -1);
			}
			else {
				printf("Trailing -d\n");
				exit(BAD_EXIT);
			}
		}
		else if (str_eq(arg, "-i")) {
			insert_flag = TRUE;
		}
		else if (str_eq(arg, "-f")) {
			if (argc--) {
				arg = *argv++;
				if (sysopen(arg) == FALSE) {
					printf("Can not open %s\n", arg);
					sysabort();
				}
				hold_flag = TRUE;
				synonyms();
				sysiclose();
				syshkill();
				ch = ' ';
				hold_flag = FALSE;
				ch_hold   = FALSE;
			}
			else {
				printf("Trailing -f\n");
				exit(BAD_EXIT);
			}
		}
		else if (str_eq(arg, "-n")) {
			/* Allow nested comments. */
			nest_flag = TRUE;
		}
		else if (str_eq(arg, "-o")) {
			usr_flag    = TRUE;
			printf_flag = FALSE;
		}
		else if (str_eq(arg, "-s")) {
			/* Define a path. */
			if (argc--) {
				arg = *argv++;
				if (n_paths >= MAX_PATHS) {
					printf("too many path names.\n");
					exit(BAD_EXIT);
				}
				else {
					p1 = str_mcat(arg, "\\");
					paths [n_paths++] = p1;
				}
			}
			else {
				printf("Trailing -s.\n");
				exit(BAD_EXIT);
			}
		}
		else if (str_eq(arg, "-t")) {
			tick_flag = TRUE;
		}
		else if(str_eq(arg, "-u")) {
			/* Suppress the initial definition of a variable. */
			if (argc--) {
				arg = *argv++;
				mst_unarg(arg);
			}
			else {
				printf("Trailing -u.\n");
				exit(BAD_EXIT);
			}
		}
		else if (str_eq(arg, "-x")) {
			slc_flag = FALSE;
		}
		else if (str_eq(arg, "-?")) {
			/* Ignore it. */
			;
		}
		else if (in == NULL) {
			in = arg;
		}
		else if (out == NULL) {
			out = arg;
		}
		else {
			printf("Extra file argument: %s\n", arg);
			exit(BAD_EXIT);
		}
	}

	/* Make sure that both file arguments were provided. */
	if (in == NULL) {
		printf("Missing input, output file arguments.\n");
		exit(BAD_EXIT);
	}
	else if (out == NULL) {
		printf("Missing output file argument.\n");
		exit(BAD_EXIT);
	}
	else if (str_eq(in, out)) {
		fatal("Can not copy input file to output file");
	}

	/*
		Set search path from the INCLUDE environment variable.
		Do this AFTER the -s paths have been entered.
	*/
	arg = getenv("INCLUDE");
	if (arg != 0) {
		strcpy(path_buf, arg);
		arg = path_buf;
		for (p = &path_buf[0]; ;) {
			if (*p != ';' && *p != '\0') {
				p++;
				continue;
			}
			if (*p == ';') {
				*p = '\0';
				p++;
			}
			if (n_paths >= MAX_PATHS) {
				printf("too many default path names.\n");
				exit(BAD_EXIT);
			}
			else {
				p1 = str_mcat(arg, "\\");
				TRACEP("main",
					printf("default path: %s\n", arg));
				paths [n_paths++] = p1;
			}
			if (*p == '\0') {
				break;
			}
			else {
				arg = p;
			}
		}
	}

	/*
		Open the output file.

		This should be done BEFORE opening the input file because
		opening the input file will cause a character to be written
		to the output file.
	*/
	if (syscreat(out) == FALSE) {
		printf("Can not open %s\n", out);
		exit(BAD_EXIT);
	}

	/* Open the input file. */
	if (sysopen(in) == FALSE) {
		printf("Can not open %s\n", in);
		sysabort();
	}

	/*
		Initialize the predefined macros (__line__ and __file__) here
		so that they can be suppressed with the -u command line option.
	*/
	mst2_init();

	/* Insert sl.h if requested. */
	if (insert_flag) {
		syssput("#include \"sl.h\"");
		sysnlput();
	}

	/*
		Start off at a new line.
		This is somewhat kludgy, because we need a GENERAL way of
		allowing preprocessor directives at the start of EVERY file.
	*/
	begin_line(TRUE);

	/* Parse the program !! */
	program();

	/* Close the output file. */
	syshflush();
	sysoclose();

	TRACE("m_stat", m_stat());
	TRACE("dump", sl_dump());

	sysend();
}

/*
	Return the next token from the input file.
	Set global variables token and t_subtype.

	Note:  Most output is created as a side effect of calling sysnext().
*/

#ifdef SHERLOCK
char bug_s1 [] = "get_t1";
char bug_s2 [] = "token %d, %s\n";
#endif

/* CAUTION: evaluate value only once! */
#define T_RETURN(value)\
	token = value;\
	TRACEPN(bug_s1, printf(bug_s2, token, pr_tok()));\
	return;

void
get_token(void)
{
	TICK("get_token");

	/* Save previous token. */
	/* Bug fix:  11/3/88 */
	if (!con_flag) {
		last_tok = token;
	}

	/* Reset first_token_in_line flag. */
	first_tok = FALSE;

	/* Output the buffer. */
	syshflush();

	/* Get the next token. */
	get_t1(TRUE);

	TRACEP("get_token",
		printf("returns: %s = %d, last_tok: %d, ch: %s\n",
			pr_tok(), token, last_tok, pr_ch(ch)));
}

/* Same as get_token but do not allow expansion. */
void
get_xtoken(void)
{
	TICK("get_xtoken");

	/* Save previous token. */
	last_tok = token;

	/* Reset first_token_in_line flag. */
	first_tok = FALSE;

	/* Output the buffer. */
	syshflush();

	/* Get the next token. */
	get_t1(FALSE);

	TRACEP("get_xtoken",
		printf("returns: %s = %d, last_tok: %d, ch: %s\n",
			pr_tok(), token, last_tok, pr_ch(ch)));
}

/*
	Kludgy buffer for held id's after a string.
*/
static bool id_flag = FALSE;
static char id_buf [MAX_SYMBOL];

void
get_t1(bool expand_flag)
{
	char *text;
	int nargs;
	int i;
	unsigned char mesgbuf [40];
	unsigned char cbuf [2];
	char old_name [100];
	bool old_mflag;

	TRACEP("get_t1", printf("(line %d)\n", t_line));

	*t_symbol = '\0';
	t_subtype = 0;

	/* Kludge required by string catenation. */
	if (id_flag) {
		id_flag = FALSE;
		strcpy(&t_symbol[0], &id_buf[0]);
		T_RETURN(ID_TOK);
	}

	/*
		We branch to the 'rescan' label whenever a construct is seen
		that does not result directly in a token being returned, i.e.,
		for macros, PP directives and whitespace.
	*/

rescan:

	TRACEP("get_t1_at_rescan", printf("ch: %s\n", pr_ch(ch)));

	/* Special case for "wide" characters and strings. */
	if (ch == 'L') {
		sysnext();
		if (isid2(ch)) {
			syspushback(ch);
			ch = 'L';
		}
	}

	if (isid1(ch)) {
		old_mflag = m_flag;
		t_id(&t_symbol[0], MAX_SYMBOL);
		token = is_reserved(t_symbol, t_length);
		if (is(NULL_TOK)) {
			token = ID_TOK;
			if (	expand_flag &&
				outer_expand(&t_symbol[0], old_mflag)
			   ) {
				goto rescan;
			}
			else {
				T_RETURN(token);
			}
		}
		else {
			T_RETURN(token);
		}
	}

	switch (ch) {

	TICK("get_token1");

	case ' ':
	case '\t':
		/* Accumulate non-leading white space. */
		hws_count = 0;
		while (ch == ' ' || ch == '\t') {
			hws_buf[hws_count++] = ch;
			sysnext();
		}
		goto rescan;

	case '\n':

		if (con_flag) {
			con_flag = FALSE;
			TRACEPN("get_token", printf("con_flag: NULL_TOK\n"));
			T_RETURN(NULL_TOK);
		}

                /* Allow user to abort here. */
		syscsts();
		sysnext();
		bump_line();

		/*	Accumulate leading white space.
			Do not change the count unless there is some.
			Set the consecutive newline count.
		*/
		cnl_count = 1;
		while (ch == ' ' || ch == '\t' || ch == '\n') {
			if (ch == '\n') {
				cnl_count++;
				bump_line();
				sysnext();
			}
			else if (!con_flag && (ch == ' ' || ch == '\t')) {
				lws_count = 0;
				while (ch == ' ' || ch == '\t') {
					lws_buf [lws_count++] = ch;
					sysnext();
				}
			}
		}
		begin_line(TRUE);
		goto rescan;

	case '#':
		error("Unexpected # ignored");
		sysnext();
		goto rescan;

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':

		T_RETURN(t_number(FALSE));

	case '.':
		sysnext();
		if (ch >= '0' && ch <= '9') {
			T_RETURN(t_number(TRUE));
		}
		else if (ch == '.') {
			sysnext();
			if (ch == '.') {
				sysnext();
				T_RETURN(DOTS3);
			}
		}
		else {
			T_RETURN(DOT_TOK);
		}

	/* -----
	case '"':
		t_string(t_symbol, MAX_SYMBOL, FALSE);
		T_RETURN(STRING_TOK);
	----- */

	case '"':

		/*	Do concatenation of string literals here.
			Note that a macro may expand into several strings.
		*/
		t_symbol[0] = '"';
		for(i = 1;;) {
			if (ch == ' ' || ch == '\t') {
				sysnext();
			}
			else if (ch == '"') {
				t_string(&t_symbol[i], MAX_SYMBOL-i, FALSE);
				i = strlen(&t_symbol[0]);
			}
			else if (isid1(ch)) {
				old_mflag = m_flag;
				t_id(&id_buf[0], MAX_SYMBOL);
				if (!outer_expand(&id_buf[0], old_mflag)) {
					/*
						Not a macro.
						Restore the id and return the
						stirng.
					*/
					id_flag = TRUE;
					break;
				}
			}
			else {
				break;
			}
		}
		t_symbol[i++] = '"';
		t_symbol[i]   = '\0';
		T_RETURN(STRING_TOK);

	case '\'':
		t_string(t_symbol, MAX_SYMBOL, FALSE);
		t_value = (long) char_val(t_symbol);
		T_RETURN(CHAR_TOK);

	/*
		We must be VERY careful about exactly when we switch from one
		input file to the next.  This is the place. 
	*/

	case END_FILE:

		if (con_flag) {
			con_flag = FALSE;
			T_RETURN(NULL_TOK);
		}

		/* Switch input streams. */
		sysiclose();
		if (t_inlevel == -1) {
			T_RETURN(EOP_TOK);
		}
		else {
			begin_line(TRUE);
			goto rescan;
		}


	case '=':	/* = or == */

		sysnext();
		if (ch == '=') {
			sysnext();
			T_RETURN(EQUAL_TOK);
		}
		else {
			T_RETURN(ASSN_TOK);
		}


	case '+':	/* + or ++ or += */

		sysnext();
		if (ch == '+') {
			sysnext();
			T_RETURN(INC_TOK);
		}
		else if (ch == '=') {
			sysnext();
			T_RETURN(PLUS_ASSN_TOK);
		}
		else {
			T_RETURN(PLUS_TOK);
		}


	case '-':	/* - or -- or -> */

		sysnext();
		if (ch == '=') {
			sysnext();
			T_RETURN(MINUS_ASSN_TOK);
		}
		else if (ch == '-') {
			sysnext();
			T_RETURN(DEC_TOK);
		}
		else if (ch == '>') {
			sysnext();
			T_RETURN(ARROW_TOK);
		}
		else {
			T_RETURN(MINUS_TOK);
		}


	case '*':	/* * or *= */

		sysnext();
		if (ch == '=') {
			sysnext();
			T_RETURN(STAR_ASSN_TOK);
		}
		else {
			T_RETURN(STAR_TOK);
		}


	case '/':	/* comment or / or /= */

		sysnext();
		if (ch == '*') {
			sysnext();
			skip_comment();
			goto rescan;
		}
		else if (slc_flag && ch == '/') {
			/* 8/3/89 MicroSoft single-line comment. */
			skip_past();
			goto rescan;
		}
		else if (ch == '=') {
			sysnext();
			T_RETURN(DIV_ASSN_TOK);
		}
		else {
			T_RETURN(DIV_TOK);
		}


	case '%':	/* % or %= */

		sysnext();
		if (ch == '=') {
			sysnext();
			T_RETURN(MOD_ASSN_TOK);
		}
		else {
			T_RETURN(MOD_TOK);
		}


	case '>':	/* > or >= or >> or >>= */

		sysnext();
		if (ch == '>') {
			sysnext();
			if (ch == '=') {
				sysnext();
				T_RETURN(RSHIFT_ASSN_TOK);
			}
			else {
				T_RETURN(RSHIFT_TOK);
			}
		}
		else if (ch == '=') {
			sysnext();
			T_RETURN(GE_TOK);
		}
		else {
			T_RETURN(GT_TOK);
		}


	case '<':	/* < or or <= or << or <<= */

		sysnext();
		if (ch == '<') {
			sysnext();
			if (ch == '=') {
				sysnext();
				T_RETURN(LSHIFT_ASSN_TOK);
			}
			else {
				T_RETURN(LSHIFT_TOK);
			}
		}
		else if (ch == '=') {
			sysnext();
			T_RETURN(LE_TOK);
		}
		else {
			T_RETURN(LT_TOK);
		}


	case '!':	/* ! or != */

		sysnext();
		if (ch == '=') {
			sysnext();
			T_RETURN(NE_TOK);
		}
		else {
			T_RETURN(NOT_TOK);
		}
			

	case '|':	/* | or |= or || */

		sysnext();
		if (ch == '=') {
			sysnext();
			T_RETURN(OR_ASSN_TOK);
		}
		else if (ch == '|') {
			sysnext();
			T_RETURN(LOR_TOK);
		}
		else {
			T_RETURN(OR_TOK);
		}


	case '&':	/* & or &= or && */

		sysnext();
		if (ch == '=') {
			sysnext();
			T_RETURN(AND_ASSN_TOK);
		}
		else if (ch == '&') {
			sysnext();
			T_RETURN(LAND_TOK);
		}
		else {
			T_RETURN(AND_TOK);
		}


	case '^':	/* ^ or ^= */

		sysnext();
		if (ch == '=') {
			sysnext();
			T_RETURN(XOR_ASSN_TOK);
		}
		else {
			T_RETURN(XOR_TOK);
		}


	case '?':	sysnext();	T_RETURN(QUESTION_TOK);
	case ':':	sysnext();	T_RETURN(COLON_TOK);

	case '~':	sysnext();	T_RETURN(TILDE_TOK);
	case ',':	sysnext();	T_RETURN(COMMA_TOK);
	case '(':	sysnext();	T_RETURN(LPAREN_TOK);
	case ')':	sysnext();	T_RETURN(RPAREN_TOK);
	case '[':	sysnext();	T_RETURN(LBRACK_TOK);
	case ']':	sysnext();	T_RETURN(RBRACK_TOK);
	case '{':	sysnext();	T_RETURN(LCURLY_TOK);
	case '}':	sysnext();	T_RETURN(RCURLY_TOK);
	case ';':	sysnext();	T_RETURN(SEMICOLON_TOK);

	default:

		strcpy(mesgbuf, "Invalid character ignored: ");
		cbuf [0] = ch;
		cbuf [1] = '\0';
		strcat(mesgbuf, cbuf);
		error(mesgbuf);
		sysnext();
		goto rescan;
	}
}

#undef T_RETURN

/*
	Do beginning of line processing.
	Look for preprocessor directives only if flag is TRUE.
*/
void
begin_line(bool flag)
{
	TRACEP("begin_line", printf("%s: %3d:\n", t_file, t_line));

	skip_ws(TRUE );

	/* PP directives are not allowed as the result of macro expansion. */
	if (!con_flag && flag && ch == '#' && !m_flag) {
		sysnext();
		do_pp();
	}

	first_tok = TRUE;
}

/*
	Do end of line processing.
*/
void
bump_line(void)
{
	t_line++;
	TRACEP("bump_line", printf("t_line = %d\n", t_line));
}

/*
	Process a synonym file.
*/
static void
synonyms(void)
{
	char buffer1 [MAX_SYMBOL];
	char buffer2 [MAX_SYMBOL];

	sysnext();
	skip_bl();
	for(;;) {
		if (ch == END_FILE) {
			return;
		}
		else if (isid1(ch)) {
			t_id(&buffer1[0], MAX_SYMBOL);
			skip_bl();
			if (isid1(ch)) {
				t_id(&buffer2[0], MAX_SYMBOL);
				set_syn(&buffer1[0], &buffer2[0]);
				TRACEP("synonym",
					printf("<%s> = <%s>\n",
						buffer1, buffer2));
				skip_bl();
				if (ch == '\n') {
					sysnext();
					skip_bl();
				}
				else {
					error("Newline expected");
					sysabort();
				}
			}
			else {
				error("Synonym expected");
				sysabort();
			}
		}
		else if (ch == ' ' || ch == '\t' || ch == '\n') {
			sysnext();
		}
		else if (ch == '#') {
			while (ch != '\n' && ch != END_FILE) {
				sysnext();
			}
		}
		else {
			error("Identifier or end-of-file expected");
			sysabort();
		}
	}
}			

/*
	Redefine the names of the Sherlock macros actually generated by SPP.
	Not every Sherlock macro is in the list--only those used by sem.c.
*/
static void
set_syn(id1, id2)
char *id1;
char *id2;
{
	if(str_eq(id1, r_bool)) {
		r_bool = str_alloc(id2);
	}
	else if(str_eq(id1, r_char)) {
		r_char = str_alloc(id2);
	}
	else if(str_eq(id1, r_double)) {
		r_double = str_alloc(id2);
	}
	else if(str_eq(id1, r_float)) {
		r_float = str_alloc(id2);
	}
	else if(str_eq(id1, r_int)) {
		r_int = str_alloc(id2);
	}
	else if(str_eq(id1, r_long)) {
		r_long = str_alloc(id2);
	}
	else if(str_eq(id1, r_ptr)) {
		r_ptr = str_alloc(id2);
	}
	else if(str_eq(id1, r_string)) {
		r_string = str_alloc(id2);
	}
	else if(str_eq(id1, r_uint)) {
		r_uint = str_alloc(id2);
	}
	else if(str_eq(id1, r_ulong)) {
		r_ulong = str_alloc(id2);
	}
	else if(str_eq(id1, r_void)) {
		r_void = str_alloc(id2);
	}
	else if(str_eq(id1, s_init)) {
		s_init = str_alloc(id2);
	}
	else if(str_eq(id1, s_parse)) {
		s_parse = str_alloc(id2);
	}
	else if(str_eq(id1, tick)) {
		tick = str_alloc(id2);
	}
	else if(str_eq(id1, tickb)) {
		tickb = str_alloc(id2);
	}
	else if(str_eq(id1, tickx)) {
		tickx = str_alloc(id2);
	}
	else if(str_eq(id1, pbtrace)) {
		pbtrace = str_alloc(id2);
	}
	else if(str_eq(id1, ptrace)) {
		ptrace = str_alloc(id2);
	}
	else {
		warn3("Synonym entry for ", id1, " has no effect");
		err_cont();
	}
}

/*
	Return TRUE if id is a current name of any Sherlock macro.
	This code is slow, but it is called only by outer_expand after
	an actual macro is seen, so its speed is not important.

	The only effect of SL_DISABLE is to cause this routine to return
	TRUE, thereby disabling generation of Sherlock macros.
*/
bool
is_sherlock(char *id)
{
	int i;
	for (i = 0;;i++) {
		if (mn_tab[i] == NULL) {
			return FALSE;
		}
		else if (str_eq(id, mn_tab[i])) {
			return TRUE;
		}
	}
}
